@@ -28,7 +28,7 @@
 #include <getopt.h>
 #include <unistd.h>
 #include <errno.h>
-#include <signal.h> 
+#include <signal.h>
 #include <pthread.h>
 #include <string.h>
 #include <pwd.h>
@@ -57,12 +57,15 @@
 int tftpd_max_thread = 100;     /* number of concurent thread allowed */
 int tftpd_timeout = 300;        /* number of second of inactivity
                                    before exiting */
-char directory[MAXLEN] = "/tftpboot/";
+char directory[MAXLEN] = "/srv/tftp/";
 int retry_timeout = S_TIMEOUT;
 
+int on = 1;
+int listen_local = 0;
+
 int tftpd_daemon = 0;           /* By default we are started by inetd */
 int tftpd_daemon_no_fork = 0;   /* For who want a false daemon mode */
-short tftpd_port = 0;           /* Port atftpd listen to */
+short tftpd_port = 69;          /* Port atftpd listen to */
 char tftpd_addr[MAXLEN] = "";   /* IP address atftpd binds to */
 
 int tftpd_cancel = 0;           /* When true, thread must exit. pthread
@@ -153,10 +156,10 @@ int main(int argc, char **argv)
      int run = 1;               /* while (run) loop */
      struct thread_data *new;   /* for allocation of new thread_data */
      int sockfd;                /* used in daemon mode */
-     struct sockaddr_in sa;     /* used in daemon mode */
-     struct servent *serv;
+     struct sockaddr_storage sa; /* used in daemon mode */
      struct passwd *user;
      struct group *group;
+     pthread_t tid;
 
 #ifdef HAVE_MTFTP
      pthread_t mtftp_thread;
@@ -177,14 +180,19 @@ int main(int argc, char **argv)
       * verify the command line arguments
       */
      if (tftpd_cmd_line_options(argc, argv) == ERR)
+     {
+          printf("Can't parse the command line options.\n");
           exit(1);
+     }
 
      /*
-      * Can't be started from the prompt without explicitely specifying 
+      * Can't be started from the prompt without explicitely specifying
       * the --daemon option.
       */
      if (isatty(0) && !(tftpd_daemon))
      {
+	  printf("Can't be started from the prompt without explicitely "
+		 "specifying the --daemon option.\n");
           tftpd_usage();
           exit(1);
      }
@@ -228,32 +236,48 @@ int main(int argc, char **argv)
                     exit(2);
           }
 
-          /* find the port */
-          if (tftpd_port == 0)
+          /* find the port; initialise sockaddr_storage structure */
+          if (strlen(tftpd_addr) > 0 || tftpd_port == 0)
           {
-               if ((serv = getservbyname("tftp", "udp")) == NULL)
+               struct addrinfo hints, *result;
+               int err;
+
+               /* look up the service and host */
+               memset(&hints, 0, sizeof(hints));
+               hints.ai_socktype = SOCK_DGRAM;
+               hints.ai_flags = AI_NUMERICHOST;
+               err = getaddrinfo(strlen(tftpd_addr) ? tftpd_addr : NULL, tftpd_port ? NULL : "tftp",
+                                 &hints, &result);
+               if (err == EAI_SERVICE)
                {
                     logger(LOG_ERR, "atftpd: udp/tftp, unknown service");
                     exit(1);
                }
-               tftpd_port = ntohs(serv->s_port);
-          }
-          /* initialise sockaddr_in structure */
-          memset(&sa, 0, sizeof(sa));
-          sa.sin_family = AF_INET;
-          sa.sin_port = htons(tftpd_port);
-          if (strlen(tftpd_addr) > 0)
-          {
-               if (inet_aton(tftpd_addr, &(sa.sin_addr)) == 0)
+               if (err || sockaddr_set_addrinfo(&sa, result))
                {
                     logger(LOG_ERR, "atftpd: invalid IP address %s", tftpd_addr);
                     exit(1);
                }
+
+               if (!tftpd_port)
+                    tftpd_port = sockaddr_get_port(&sa);
+               else {
+                    sa.ss_family = AF_INET;
+                    sockaddr_set_port(&sa, tftpd_port);
+               }
+
+               freeaddrinfo(result);
           }
-          else
-               sa.sin_addr.s_addr = htonl(INADDR_ANY);
+
+          if (strlen(tftpd_addr) == 0)
+          {
+               memset(&sa, 0, sizeof(sa));
+               sa.ss_family = AF_INET;
+          }
+          sockaddr_set_port(&sa, tftpd_port);
+
           /* open the socket */
-          if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) == 0)
+          if ((sockfd = socket(sa.ss_family, SOCK_DGRAM, 0)) == 0)
           {
                logger(LOG_ERR, "atftpd: can't open socket");
                exit(1);
@@ -287,24 +311,46 @@ int main(int argc, char **argv)
           if (pidfile)
           {
                if (tftpd_pid_file(pidfile, 1) != OK)
+               {
+                    logger(LOG_ERR,
+                           "atftpd: can't write our pid file: %s.",
+                           pidfile);
                     exit(1);
+               }
                /* to be able to remove it later */
-               chown(pidfile, user->pw_uid, group->gr_gid);
+               if (chown(pidfile, user->pw_uid, group->gr_gid) != OK) {
+	            logger(LOG_ERR,
+		           "atftpd: failed to chown our pid file %s to owner %s.%s.",
+                           pidfile, user_name, group_name);
+                    exit(1);
+	       }
           }
 
-          setgid(group->gr_gid);
-          setuid(user->pw_uid);
+	  if (setgid(group->gr_gid) != OK) {
+	      logger(LOG_ERR,
+		      "atftpd: failed to setgid to group %d (%s).",
+		      group->gr_gid, group_name);
+	      exit(1);
+	  }
+	  if (setuid(user->pw_uid) != OK) {
+	      logger(LOG_ERR,
+		      "atftpd: failed to setuid to user %d (%s).",
+		      user->pw_uid, user_name);
+	      exit(1);
+	  }
 
           /* Reopen log file now that we changed user, and that we've
            * open and dup2 the socket. */
           open_logger("atftpd", log_file, logging_level);
      }
 
+#if defined(SOL_IP) && defined(IP_PKTINFO)
      /* We need to retieve some information from incomming packets */
      if (setsockopt(0, SOL_IP, IP_PKTINFO, &one, sizeof(one)) != 0)
      {
           logger(LOG_WARNING, "Failed to set socket option: %s", strerror(errno));
      }
+#endif
 
      /* save main thread ID for proper signal handling */
      main_thread_id = pthread_self();
@@ -343,7 +389,7 @@ int main(int argc, char **argv)
                          logger(LOG_ERR, "Failed to start mtftp server thread");
                          tftpd_mtftp_clean(mtftp_data);
                          mtftp_data = NULL;
-                    } 
+                    }
               }
           }
      }
@@ -366,7 +412,7 @@ int main(int argc, char **argv)
              is done reading the request. */
           pthread_mutex_lock(&stdin_mutex);
 
-#ifdef RATE_CONTROL          
+#ifdef RATE_CONTROL
           /* if we want to implement rate control, we sleep some time here
              so we cannot exceed the allowed thread/sec. */
           if (rate > 0)
@@ -387,10 +433,18 @@ int main(int argc, char **argv)
              packets */
           if (!tftpd_cancel)
           {
+               int rv;
+
                if ((tftpd_timeout == 0) || (tftpd_daemon))
-                    select(FD_SETSIZE, &rfds, NULL, NULL, NULL);
+                    rv = select(FD_SETSIZE, &rfds, NULL, NULL, NULL);
                else
-                    select(FD_SETSIZE, &rfds, NULL, NULL, &tv);
+                    rv = select(FD_SETSIZE, &rfds, NULL, NULL, &tv);
+               if (rv < 0) {
+                    logger(LOG_ERR, "%s: %d: select: %s",
+                           __FILE__, __LINE__, strerror(errno));
+                    /* Clear the bits, they are undefined! */
+                    FD_ZERO(&rfds);
+	       }
           }
 
 #ifdef RATE_CONTROL
@@ -429,7 +483,7 @@ int main(int argc, char **argv)
                new->data_buffer_size = SEGSIZE + 4;
 
                /* Allocate memory for tftp option structure. */
-               if ((new->tftp_options = 
+               if ((new->tftp_options =
                     malloc(sizeof(tftp_default_options))) == NULL)
                {
                     logger(LOG_ERR, "%s: %d: Memory allocation failed",
@@ -464,9 +518,9 @@ int main(int argc, char **argv)
                }
                new->client_info->done = 0;
                new->client_info->next = NULL;
-               
+
                /* Start a new server thread. */
-               if (pthread_create(&new->tid, NULL, tftpd_receive_request,
+               if (pthread_create(&tid, NULL, tftpd_receive_request,
                                   (void *)new) != 0)
                {
                     logger(LOG_ERR, "Failed to start new thread");
@@ -477,7 +531,7 @@ int main(int argc, char **argv)
           else
           {
                pthread_mutex_unlock(&stdin_mutex);
-               
+
                /* Either select return after timeout of we've been killed. In the first case
                   we wait for server thread to finish, in the other we kill them */
                if (tftpd_cancel)
@@ -516,7 +570,7 @@ int main(int argc, char **argv)
      /* stop collecting stats and print them*/
      stats_end();
      stats_print();
-     
+
 #ifdef HAVE_PCRE
      /* remove allocated memory for tftpd_pcre */
      if (pcre_top)
@@ -540,7 +594,7 @@ int main(int argc, char **argv)
      exit(0);
 }
 
-/* 
+/*
  * This function handles the initial connection with a client. It reads
  * the request from stdin and then release the stdin_mutex, so the main
  * thread may listen for new clients. After that, we process options and
@@ -551,29 +605,32 @@ int main(int argc, char **argv)
 void *tftpd_receive_request(void *arg)
 {
      struct thread_data *data = (struct thread_data *)arg;
-     
+
      int retval;                /* hold return value for testing */
      int data_size;             /* returned size by recvfrom */
      char string[MAXLEN];       /* hold the string we pass to the logger */
      int num_of_threads;
      int abort = 0;             /* 1 if we need to abort because the maximum
-                                   number of threads have been reached*/ 
-     struct sockaddr_in to;     /* destination of client's packet */
-     socklen_t len = sizeof(struct sockaddr);
+                                   number of threads have been reached*/
+     struct sockaddr_storage to; /* destination of client's packet */
+     socklen_t len = sizeof(to);
 
-#ifdef HAVE_WRAP
-     char client_addr[16];
-#endif
+     char addr_str[SOCKADDR_PRINT_ADDR_LEN];
 
      /* Detach ourself. That way the main thread does not have to
       * wait for us with pthread_join. */
-     pthread_detach(pthread_self());
+     data->tid = pthread_self();
+     pthread_detach(data->tid);
 
      /* Read the first packet from stdin. */
-     data_size = data->data_buffer_size;     
+     data_size = data->data_buffer_size;
      retval = tftp_get_packet(0, -1, NULL, &data->client_info->client, NULL,
                               &to, data->timeout, &data_size,
                               data->data_buffer);
+     if (retval == ERR) {
+          logger(LOG_NOTICE, "Invalid request in 1st packet");
+          abort = 1;
+     }
 
      /* now unlock stdin */
      pthread_mutex_unlock(&stdin_mutex);
@@ -587,16 +644,21 @@ void *tftpd_receive_request(void *arg)
      }
 
 #ifdef HAVE_WRAP
+     if (!abort && !sockaddr_family_supported(&data->client_info->client))
+     {
+          logger(LOG_ERR, "Connection from unsupported network address family refused");
+          abort = 1;
+     }
      if (!abort)
      {
           /* Verify the client has access. We don't look for the name but
              rely only on the IP address for that. */
-          inet_ntop(AF_INET, &data->client_info->client.sin_addr,
-                    client_addr, sizeof(client_addr));
-          if (hosts_ctl("in.tftpd", STRING_UNKNOWN, client_addr,
+          sockaddr_print_addr(&data->client_info->client,
+                              addr_str, sizeof(addr_str));
+          if (hosts_ctl("in.tftpd", STRING_UNKNOWN, addr_str,
                         STRING_UNKNOWN) == 0)
           {
-               logger(LOG_ERR, "Connection refused from %s", client_addr);
+               logger(LOG_ERR, "Connection refused from %s", addr_str);
                abort = 1;
           }
      }
@@ -612,9 +674,33 @@ void *tftpd_receive_request(void *arg)
      else
      {
           /* open a socket for client communication */
-          data->sockfd = socket(PF_INET, SOCK_DGRAM, 0);
-          to.sin_family = AF_INET;
-          to.sin_port = 0;
+          data->sockfd = socket(data->client_info->client.ss_family,
+                                SOCK_DGRAM, 0);
+          /*memset(&to, 0, sizeof(to));*/
+          /* PSz 11 Aug 2011  http://bugs.debian.org/613582
+           * Do not nullify "to", preserve IP address from tftp_get_packet().
+           * Only set port to 0, as we used to in v6.
+           * (Why set ss_family, was not it right already??)
+           */
+          to.ss_family = data->client_info->client.ss_family;
+          sockaddr_set_port(&to, 0);
+          /* Force socket to listen on local address. Do not listen on broadcast address 255.255.255.255.
+             If the socket listens on the broadcast address, Linux tells the remote client the port
+             is unreachable. This happens even if SO_BROADCAST is set in setsockopt for this socket.
+             I was unable to find a kernel option or /proc/sys flag to make the kernel pay attention to
+             these requests, so the workaround is to force listening on the local address. */
+          if (listen_local == 1)
+          {
+               logger(LOG_INFO, "forcing socket to listen on local address");
+               if (setsockopt(data->sockfd, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on)) != 0) {
+                  logger(LOG_ERR, "setsockopt: %s", strerror(errno));
+               }
+          }
+          else
+          {
+               logger(LOG_INFO, "socket may listen on any address, including broadcast");
+          }
+
           if (data->sockfd > 0)
           {
                /* bind the socket to the interface */
@@ -624,13 +710,16 @@ void *tftpd_receive_request(void *arg)
                     retval = ABORT;
                }
                /* read back assigned port */
-               len = sizeof(struct sockaddr);
+               len = sizeof(to);
                if (getsockname(data->sockfd, (struct sockaddr *)&to, &len) == -1)
                {
                     logger(LOG_ERR, "getsockname: %s", strerror(errno));
                     retval = ABORT;
                }
                /* connect the socket, faster for kernel operation */
+               /* this is not a good idea on FreeBSD, because sendto() cannot
+                  be used on a connected datagram socket */
+#if !defined(__FreeBSD_kernel__)
                if (connect(data->sockfd,
                            (struct sockaddr *)&data->client_info->client,
                            sizeof(data->client_info->client)) == -1)
@@ -638,9 +727,11 @@ void *tftpd_receive_request(void *arg)
                     logger(LOG_ERR, "connect: %s", strerror(errno));
                     retval = ABORT;
                }
+#endif
                logger(LOG_DEBUG, "Creating new socket: %s:%d",
-                      inet_ntoa(to.sin_addr), ntohs(to.sin_port));
-               
+                      sockaddr_print_addr(&to, addr_str, sizeof(addr_str)),
+                      sockaddr_get_port(&to));
+
                /* read options from request */
                opt_parse_request(data->data_buffer, data_size,
                                  data->tftp_options);
@@ -657,8 +748,9 @@ void *tftpd_receive_request(void *arg)
           case GET_RRQ:
                logger(LOG_NOTICE, "Serving %s to %s:%d",
                       data->tftp_options[OPT_FILENAME].value,
-                      inet_ntoa(data->client_info->client.sin_addr),
-                      ntohs(data->client_info->client.sin_port));
+                      sockaddr_print_addr(&data->client_info->client,
+                                          addr_str, sizeof(addr_str)),
+                      sockaddr_get_port(&data->client_info->client));
                if (data->trace)
                     logger(LOG_DEBUG, "received RRQ <%s>", string);
                if (tftpd_send_file(data) == OK)
@@ -668,7 +760,8 @@ void *tftpd_receive_request(void *arg)
                break;
           case GET_WRQ:
                logger(LOG_NOTICE, "Fetching from %s to %s",
-                      inet_ntoa(data->client_info->client.sin_addr),
+                      sockaddr_print_addr(&data->client_info->client,
+                                          addr_str, sizeof(addr_str)),
                       data->tftp_options[OPT_FILENAME].value);
                if (data->trace)
                     logger(LOG_DEBUG, "received WRQ <%s>", string);
@@ -693,7 +786,9 @@ void *tftpd_receive_request(void *arg)
                break;
           default:
                logger(LOG_NOTICE, "Invalid request <%d> from %s",
-                      retval, inet_ntoa(data->client_info->client.sin_addr));
+                      retval,
+                      sockaddr_print_addr(&data->client_info->client,
+                                          addr_str, sizeof(addr_str)));
                tftp_send_error(data->sockfd, &data->client_info->client,
                                EBADOP, data->data_buffer, data->data_buffer_size);
                if (data->trace)
@@ -701,7 +796,7 @@ void *tftpd_receive_request(void *arg)
                            tftp_errmsg[EBADOP]);
                stats_err_locked();
           }
-     }  
+     }
 
      /* make sure all data is sent to the network */
      if (data->sockfd)
@@ -733,7 +828,7 @@ void *tftpd_receive_request(void *arg)
 
      /* free the thread structure */
      free(data);
-     
+
      logger(LOG_INFO, "Server thread exiting");
      pthread_exit(NULL);
 }
@@ -811,6 +906,7 @@ int tftpd_cmd_line_options(int argc, char **argv)
           { "no-multicast", 0, NULL, 'M' },
           { "logfile", 1, NULL, 'L' },
           { "pidfile", 1, NULL, 'I'},
+          { "listen-local", 0, NULL, 'F'},
           { "daemon", 0, NULL, 'D' },
           { "no-fork", 0, NULL, 'N'},
           { "user", 1, NULL, 'U'},
@@ -837,7 +933,7 @@ int tftpd_cmd_line_options(int argc, char **argv)
           { "help", 0, NULL, 'h' },
           { 0, 0, 0, 0 }
      };
-     
+
      while ((c = getopt_long(argc, argv, "t:r:m:v::Vh",
                              options, NULL)) != EOF)
      {
@@ -862,6 +958,9 @@ int tftpd_cmd_line_options(int argc, char **argv)
                     logging_level = atoi(optarg);
                else
                     logging_level++;
+
+	       if (logging_level > LOG_DEBUG)
+		   logging_level = LOG_DEBUG;
 #ifdef HAVE_WRAP
                allow_severity = logging_level;
                deny_severity = logging_level;
@@ -888,6 +987,9 @@ int tftpd_cmd_line_options(int argc, char **argv)
           case 'I':
                pidfile = strdup(optarg);
                break;
+          case 'F':
+               listen_local = 1;
+               break;
           case 'D':
                tftpd_daemon = 1;
                break;
@@ -900,7 +1002,7 @@ int tftpd_cmd_line_options(int argc, char **argv)
                     Strncpy(user_name, tmp, MAXLEN);
                tmp = strtok(NULL, "");
                if (tmp != NULL)
-                    Strncpy(group_name, optarg, MAXLEN);
+                    Strncpy(group_name, tmp, MAXLEN);
                break;
           case 'G':
                Strncpy(group_name, optarg, MAXLEN);
@@ -908,7 +1010,7 @@ int tftpd_cmd_line_options(int argc, char **argv)
           case 'P':
                tftpd_port = (short)atoi(optarg);
                break;
-          case 'A': 
+          case 'A':
                Strncpy(tftpd_addr, optarg, MAXLEN);
                break;
           case OPT_MCAST_TTL:
@@ -964,7 +1066,7 @@ int tftpd_cmd_line_options(int argc, char **argv)
                break;
 #ifdef HAVE_MTFTP
           case OPT_MTFTP:
-               Strncpy(mtftp_file, optarg, MAXLEN);         
+               Strncpy(mtftp_file, optarg, MAXLEN);
                break;
           case OPT_MTFTP_PORT:
                mtftp_sport = atoi(optarg);
@@ -981,7 +1083,7 @@ int tftpd_cmd_line_options(int argc, char **argv)
                break;
           }
      }
-          
+
      /* verify that only one arguement is left */
      if (optind < argc)
           Strncpy(directory, argv[optind], MAXLEN);
@@ -990,7 +1092,10 @@ int tftpd_cmd_line_options(int argc, char **argv)
           strcat(directory, "/");
      /* build multicast address/port range */
      if (tftpd_mcast_parse_opt(mcast_addr, mcast_port) != OK)
+     {
+          printf("Failed to parse multicast options\n");
           exit(1);
+     }
      return OK;
 }
 
@@ -1015,6 +1120,10 @@ void tftpd_log_options(void)
      logger(LOG_INFO, "  log file: %s", (log_file==NULL) ? "syslog":log_file);
      if (pidfile)
           logger(LOG_INFO, "  pid file: %s", pidfile);
+     if (listen_local == 1)
+          logger(LOG_INFO, "  forcing to listen on local interfaces: on.");
+     else
+          logger(LOG_INFO, "  not forcing to listen on local interfaces.");
      if (tftpd_daemon == 1)
           logger(LOG_INFO, "  server timeout: Not used");
      else
@@ -1054,7 +1163,7 @@ void tftpd_log_options(void)
           logger(LOG_INFO, "  --no-source-port-checking turned on");
 }
 
-/* 
+/*
  *
  */
 int tftpd_pid_file(char *file, int action)
@@ -1077,7 +1186,7 @@ int tftpd_pid_file(char *file, int action)
                return ERR;
           }
           /* write it */
-          pid = getpid();     
+          pid = getpid();
           fprintf(fp, "%d\n", pid);
           fclose(fp);
           return OK;
@@ -1111,11 +1220,12 @@ void tftpd_usage(void)
             " output messages\n"
             "  --trace                    : log all sent and received packets\n"
             "  --no-timeout               : disable 'timeout' from RFC2349\n"
-            "  --no-tisize                : disable 'tsize' from RFC2349\n"
+            "  --no-tsize                 : disable 'tsize' from RFC2349\n"
             "  --no-blksize               : disable 'blksize' from RFC2348\n"
             "  --no-multicast             : disable 'multicast' from RFC2090\n"
-            "  --logfile <file>           : logfile to log logs to ;-)\n"
+            "  --logfile <file>           : logfile to log logs to ;-) (use - for stdout)\n"
             "  --pidfile <file>           : write PID to this file\n"
+            "  --listen-local             : force listen on local network address\n"
             "  --daemon                   : run atftpd standalone (no inetd)\n"
             "  --no-fork                  : run as a daemon, don't fork\n"
             "  --user <user[.group]>      : default is nobody\n"
